#include <WiFi.h>
#include <HTTPClient.h>
#include <ArduinoJson.h>
#include <freertos/FreeRTOS.h>
#include <freertos/task.h>
#include <freertos/queue.h>

#define RXD2 17
#define TXD2 16

HardwareSerial nextionSerial(2);  // Use Serial2
// Global chart configuration with fixed division size
struct ChartConfig {
  float topValue;         // current + 120
  float upperMidValue1;   // current + 80
  float upperMidValue2;   // current + 40
  float lowerMidValue1;   // current - 40
  float lowerMidValue2;   // current - 80
  float bottomValue;      // current - 120
  String currentCoin;
};

// Updated constants for 6 divisions
const float DIVISION_SIZE = 40.0f;  // Each division = 40 units
const int CHART_HEIGHT = 240;       // 6 divisions × 40 = 240 range
const int DIVISION_COUNT = 6;       // Total divisions
// Replace with your network credentials
const char *ssid = "JioFiber";
const char *password = "your password";

// RTOS handles
TaskHandle_t SerialTaskHandle = NULL;
TaskHandle_t FetchTaskHandle = NULL;
QueueHandle_t commandQueue = NULL;

// Debugging flags
const bool DEBUG_NEXTION_TX = false;  // Print all data sent to Nextion
const bool DEBUG_NEXTION_RX = false;  // Print all data received from Nextion

// Binance API endpoint
const char *binance_api_base = "https://api.binance.com";

// Time period options
enum TimePeriod {
  PERIOD_1M,   // 1 minute
  PERIOD_5M,   // 5 minutes
  PERIOD_15M,  // 15 minutes
  PERIOD_1H,   // 1 hour
  PERIOD_4H,   // 4 hours
  PERIOD_24H,  // 24 hours
  PERIOD_7D    // 7 days
};

// Command structure for inter-task communication
struct Command {
  enum Type {
    PERIOD_CHANGE,
    CHART_COIN_CHANGE
  };
  Type type;
  TimePeriod period;
  String chartCoin;
};

// Global period setting
volatile TimePeriod globalPeriod = PERIOD_24H;  // Default period
String currentChartCoin = "BTCUSDT";            // Default chart coin
float chartMinValue = 0;
float chartMaxValue = 0;
float chartRange = 1;  // Prevent division by zero
bool shouldUpdateChart = false;

// Structure to hold coin data
struct CoinData {
  String symbol;
  float currentPrice;
  float priceChangePercent;
  float periodHigh;
  float periodLow;
  float prevPrice;  // To store previous price for color change comparison
};

// Array of coins to track with their display components
const struct {
  String symbol;
  String priceComp;
  String changeComp;
  String highComp;
  String lowComp;
} coinDisplayMap[] = {
  { "BTCUSDT", "t6", "t7", "t8", "t9" },
  { "ETHUSDT", "t11", "t12", "t13", "t14" },
  { "BNBUSDT", "t16", "t17", "t18", "t19" },
  { "SOLUSDT", "t21", "t22", "t23", "t24" },
  { "XRPUSDT", "t26", "t27", "t28", "t29" },
  { "ADAUSDT", "t31", "t32", "t33", "t34" },
  { "DOGEUSDT", "t36", "t37", "t38", "t39" },
  { "AVAXUSDT", "t41", "t42", "t43", "t44" },
  { "DOTUSDT", "t46", "t47", "t48", "t49" },
  { "LINKUSDT", "t51", "t52", "t53", "t54" },
  { "MATICUSDT", "t56", "t57", "t58", "t59" },
  { "LTCUSDT", "t61", "t62", "t63", "t64" },
  { "TRXUSDT", "t66", "t67", "t68", "t69" }
};
const uint8_t coinCount = 13;

// Array to store the coin data
CoinData coinData[coinCount];
ChartConfig chartConfig; 
// Function prototypes
void connectToWiFi();
String getIntervalString(TimePeriod period);
void setGlobalPeriod(TimePeriod period);
void fetchCoinData(uint8_t index);
String periodToString(TimePeriod period);
void parseNextionSerial();
void sendToNextion(const String &cmd);
void updateNextionDisplay(uint8_t index);
void printCoinData(uint8_t index);
void updateChart();
void updateYAxisLabels(float currentPrice);
void processChartCommand(const String &symbol);
void processPeriodCommand(const String &periodStr);

// RTOS Task functions
void serialTask(void *pvParameters);
void fetchTask(void *pvParameters);

void setup() {
  Serial.begin(115200);
  nextionSerial.begin(115200, SERIAL_8N1, RXD2, TXD2);

  // Create command queue
  commandQueue = xQueueCreate(10, sizeof(Command *));

  // Create tasks
  xTaskCreatePinnedToCore(
    serialTask,         // Task function
    "SerialTask",       // Name
    10000,              // Stack size
    NULL,               // Parameters
    1,                  // Priority
    &SerialTaskHandle,  // Task handle
    0                   // Core (0 or 1)
  );

  xTaskCreatePinnedToCore(
    fetchTask,         // Task function
    "FetchTask",       // Name
    15000,             // Stack size
    NULL,              // Parameters
    1,                 // Priority
    &FetchTaskHandle,  // Task handle
    1                  // Core (0 or 1)
  );

  // Initialize coin data
  for (uint8_t i = 0; i < coinCount; i++) {
    coinData[i].symbol = coinDisplayMap[i].symbol;
    coinData[i].prevPrice = 0;
  }

  connectToWiFi();

  // Set default period
  setGlobalPeriod(PERIOD_24H);

  // Initialize chart range with default values
  chartMinValue = 0;
  chartMaxValue = 100000;
  chartRange = chartMaxValue - chartMinValue;

  // Delete the setup/loop task since we're using RTOS tasks
  vTaskDelete(NULL);
}

void loop() {
  // Empty - everything is handled in tasks
}

// Serial Task - Handles all Nextion communication
void serialTask(void *pvParameters) {
  for (;;) {
    parseNextionSerial();
    vTaskDelay(2 / portTICK_PERIOD_MS);  // Small delay to prevent watchdog trigger
  }
}

// Fetch Task - Handles all data fetching and display updates
void fetchTask(void *pvParameters) {
  TickType_t xLastWakeTime = xTaskGetTickCount();
  const TickType_t xFrequency = pdMS_TO_TICKS(500);  // 6 second interval

  for (;;) {
    // Process any pending commands from the queue
    Command *cmd;
    while (xQueueReceive(commandQueue, &cmd, 0)) {
      switch (cmd->type) {
        case Command::PERIOD_CHANGE:
          setGlobalPeriod(cmd->period);
          break;
        case Command::CHART_COIN_CHANGE:
          currentChartCoin = cmd->chartCoin;
          Serial.println("Chart coin changed to: " + currentChartCoin);
          break;
      }
      delete cmd;
    }

    if (WiFi.status() == WL_CONNECTED) {
      // Send connected WiFi name to g0
      sendToNextion("g0.txt=\"" + WiFi.SSID() + "\"");

      Serial.println("\nFetching data...");
      Serial.println("Current Period: " + periodToString(globalPeriod));

      // Fetch and display data for each coin
      for (uint8_t i = 0; i < coinCount; i++) {
        fetchCoinData(i);
       // printCoinData(i);
        updateNextionDisplay(i);
        vTaskDelay(5 / portTICK_PERIOD_MS);  // Small delay between requests
      }

      // Update chart for the selected coin only if requested
      if (shouldUpdateChart) {
        updateChart();
        // updateYAxisLabels(coinData[i].currentPrice);
        shouldUpdateChart = false;
      }
    } else {
      Serial.println("WiFi not connected, skipping data fetch");
    }

    vTaskDelayUntil(&xLastWakeTime, xFrequency);
  }
}

void connectToWiFi() {
  Serial.print("Connecting to Wi-Fi");
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("\nConnected to Wi-Fi");
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());
}

String getIntervalString(TimePeriod period) {
  switch (period) {
    case PERIOD_1M: return "1m";
    case PERIOD_5M: return "5m";
    case PERIOD_15M: return "15m";
    case PERIOD_1H: return "1h";
    case PERIOD_4H: return "4h";
    case PERIOD_24H: return "1d";
    case PERIOD_7D: return "1w";
    default: return "1d";
  }
}

void setGlobalPeriod(TimePeriod period) {
  globalPeriod = period;
  Serial.println("Global period set to: " + getIntervalString(period));

  // Update Nextion display with new period
  String cmd = "n0.val=" + String((int)period);
  sendToNextion(cmd);
}

void fetchCoinData(uint8_t index) {
  if (WiFi.status() != WL_CONNECTED) {
    Serial.println("Wi-Fi not connected.");
    return;
  }

  HTTPClient http;
  String url;

  // Use different API endpoints based on period
  if (globalPeriod == PERIOD_24H) {
    // Use 24hr ticker for daily data
    url = String(binance_api_base) + "/api/v3/ticker/24hr?symbol=" + coinData[index].symbol;
  } else {
    // Use klines for custom periods
    url = String(binance_api_base) + "/api/v3/klines?symbol=" + coinData[index].symbol + "&interval=" + getIntervalString(globalPeriod) + "&limit=1";
  }

  http.begin(url);
  int httpCode = http.GET();

  if (httpCode == 200) {
    String payload = http.getString();
    DynamicJsonDocument doc(1024);
    deserializeJson(doc, payload);

    // Store previous price before updating
    coinData[index].prevPrice = coinData[index].currentPrice;

    if (globalPeriod == PERIOD_24H) {
      // Parse 24hr ticker data
      coinData[index].currentPrice = doc["lastPrice"].as<float>();
      coinData[index].priceChangePercent = doc["priceChangePercent"].as<float>();
      coinData[index].periodHigh = doc["highPrice"].as<float>();
      coinData[index].periodLow = doc["lowPrice"].as<float>();
    } else {
      // Parse klines data
      JsonArray kline = doc[0];
      coinData[index].currentPrice = kline[4].as<float>();  // Close price
      coinData[index].periodHigh = kline[2].as<float>();    // High price
      coinData[index].periodLow = kline[3].as<float>();     // Low price

      // Calculate percentage change (close - open) / open
      float openPrice = kline[1].as<float>();
      coinData[index].priceChangePercent = ((coinData[index].currentPrice - openPrice) / openPrice) * 100;
    }

    // Update chart range if this is the chart coin
    if (coinData[index].symbol == currentChartCoin) {
      // Adjust chart range to be ±10% of current price
      chartMinValue = coinData[index].currentPrice * 0.9;
      chartMaxValue = coinData[index].currentPrice * 1.1;
      chartRange = chartMaxValue - chartMinValue;
    }
  } else {
    Serial.println("Failed to fetch data for " + coinData[index].symbol + ". Error: " + http.errorToString(httpCode));
  }

  http.end();
}

String periodToString(TimePeriod period) {
  switch (period) {
    case PERIOD_1M: return "1 minute";
    case PERIOD_5M: return "5 minutes";
    case PERIOD_15M: return "15 minutes";
    case PERIOD_1H: return "1 hour";
    case PERIOD_4H: return "4 hours";
    case PERIOD_24H: return "24 hours";
    case PERIOD_7D: return "7 days";
    default: return "Unknown";
  }
}

void parseNextionSerial() {
  static String messageBuffer;
  static bool inMessage = false;
  static char startChar = 0;
  static bool inHexMessage = false;
  static uint8_t hexBytes[3];
  static uint8_t hexIndex = 0;

  while (nextionSerial.available()) {
    uint8_t c = nextionSerial.read();

    if (DEBUG_NEXTION_RX) {
      Serial.print("[RX] Raw byte: 0x");
      if (c < 0x10) Serial.print("0");
      Serial.print(c, HEX);
      Serial.print(" (");
      if (isPrintable(c)) Serial.print((char)c);
      else Serial.print(".");
      Serial.println(")");
    }

    // Check for hex message (0x7DF)
    if (!inHexMessage && c == 0x7D) {
      inHexMessage = true;
      hexIndex = 0;
      hexBytes[hexIndex++] = c;
      continue;
    }

    if (inHexMessage) {
      hexBytes[hexIndex++] = c;
      if (hexIndex == 3) {
        inHexMessage = false;
        if (hexBytes[0] == 0x7D && hexBytes[1] == 0x00 && hexBytes[2] == 0xFF) {
          // Received 0x7D 0x00 0xFF (0x7DF)
          shouldUpdateChart = true;
          Serial.println("Received chart update request (0x7DF)");
        }
      }
      continue;
    }

    // Detect message start
    if (!inMessage && (c == '#' || c == '!')) {
      inMessage = true;
      startChar = c;
      messageBuffer = "";
      continue;
    }

    // Handle message collection
    if (inMessage) {
      // Check for message termination
      if ((startChar == '#' && c == '$') || (startChar == '!' && c == '@')) {
        // Process complete message
        if (startChar == '#') {
          processChartCommand(messageBuffer);
        } else if (startChar == '!') {
          processPeriodCommand(messageBuffer);
        }
        inMessage = false;
      } else {
        // Add character to buffer
        messageBuffer += (char)c;
      }
    }
  }
}

void processChartCommand(const String &symbol) {
  // Validate the symbol exists in our list
 String trimmedStr = symbol;
  trimmedStr.trim();
  for (uint8_t i = 0; i < coinCount; i++) {
    if (coinDisplayMap[i].symbol == trimmedStr) {
      Command *cmd = new Command();
      cmd->type = Command::CHART_COIN_CHANGE;
      cmd->chartCoin =trimmedStr;
      xQueueSend(commandQueue, &cmd, portMAX_DELAY);

      Serial.println("Chart coin change request: " + trimmedStr);
      return;
    }
  }
  Serial.println("Invalid chart coin: " + trimmedStr);
}

void processPeriodCommand(const String &periodStr) {
  // First trim any whitespace from the string
  String trimmedStr = periodStr;
  trimmedStr.trim();
  
  TimePeriod period;
  if (trimmedStr == "PERIOD_1M") period = PERIOD_1M;
  else if (trimmedStr == "PERIOD_5M") period = PERIOD_5M;
  else if (trimmedStr == "PERIOD_15M") period = PERIOD_15M;
  else if (trimmedStr == "PERIOD_1H") period = PERIOD_1H;
  else if (trimmedStr == "PERIOD_4H") period = PERIOD_4H;
  else if (trimmedStr == "PERIOD_24H") period = PERIOD_24H;
  else if (trimmedStr == "PERIOD_7D") period = PERIOD_7D;
  else {
    Serial.println("Invalid period: '" + trimmedStr + "'");
    return;
  }

  Command *cmd = new Command();
  cmd->type = Command::PERIOD_CHANGE;
  cmd->period = period;
  xQueueSend(commandQueue, &cmd, portMAX_DELAY);

  Serial.println("Period changed to: " + trimmedStr);
}

void sendToNextion(const String &cmd) {
  if (DEBUG_NEXTION_TX) {
    Serial.print("[TX] Sending to Nextion: ");
    Serial.println(cmd);
  }

  nextionSerial.print(cmd);
  nextionSerial.write(0xFF);
  nextionSerial.write(0xFF);
  nextionSerial.write(0xFF);
}

void updateNextionDisplay(uint8_t index) {
  const CoinData &data = coinData[index];
  const auto &display = coinDisplayMap[index];

  // Update price (t6, t11, t16, etc.)
  String priceCmd = display.priceComp + ".txt=\"" + String(data.currentPrice, 6) + "\"";
  sendToNextion(priceCmd);

  // Update price change (t7, t12, t17, etc.)
  String changeCmd = display.changeComp + ".txt=\"" + String(data.priceChangePercent, 2) + "%\"";
  sendToNextion(changeCmd);

  // Update high (t8, t13, t18, etc.)
  String highCmd = display.highComp + ".txt=\"" + String(data.periodHigh, 6) + "\"";
  sendToNextion(highCmd);

  // Update low (t9, t14, t19, etc.)
  String lowCmd = display.lowComp + ".txt=\"" + String(data.periodLow, 6) + "\"";
  sendToNextion(lowCmd);

  // Set colors based on values
  // Price change color (red for negative, green 2016 for positive)
  String changeColorCmd = display.changeComp + ".pco=" + String((data.priceChangePercent < 0) ? "63488" : "2016");
  sendToNextion(changeColorCmd);

  // Price color (red if current < previous, green otherwise)
  if (data.prevPrice > 0) {  // Only compare if we have previous data
    String priceColorCmd = display.priceComp + ".pco=" + String((data.currentPrice < data.prevPrice) ? "63488" : "2016");
    sendToNextion(priceColorCmd);
  }
}




void updateChart() {
  for (uint8_t i = 0; i < coinCount; i++) {
    if (coinData[i].symbol == currentChartCoin) {
      float currentPrice = coinData[i].currentPrice;
      
      // Check if we need to recalculate Y-axis
      bool needsUpdate = false;
      if (chartConfig.currentCoin != currentChartCoin) {
        needsUpdate = true;
      } 
      else if (currentPrice > chartConfig.topValue || currentPrice < chartConfig.bottomValue) {
        needsUpdate = true;
      }

      // Recalculate Y-axis if needed
      if (needsUpdate) {
        // Set fixed divisions based on current price
        chartConfig.topValue = currentPrice + DIVISION_SIZE * 3;    // current + 120
        chartConfig.upperMidValue1 = currentPrice + DIVISION_SIZE * 2; // current + 80
        chartConfig.upperMidValue2 = currentPrice + DIVISION_SIZE;   // current + 40
        chartConfig.lowerMidValue1 = currentPrice - DIVISION_SIZE;   // current - 40
        chartConfig.lowerMidValue2 = currentPrice - DIVISION_SIZE * 2; // current - 80
        chartConfig.bottomValue = currentPrice - DIVISION_SIZE * 3; // current - 120
        chartConfig.currentCoin = currentChartCoin;
        
        updateYAxisLabels();
      }

      // Scale current price to 0-240 range
      int scaledValue = scalePriceToChart(currentPrice);
      String cmd = "add 86,0," + String(scaledValue);
      sendToNextion(cmd);

      // Debug output
      if (DEBUG_NEXTION_TX) {
        Serial.print(currentChartCoin);
        Serial.print(" Price:");
        Serial.print(currentPrice, 2);
        Serial.print(" Scaled:");
        Serial.print(scaledValue);
        Serial.print(" Range:");
        Serial.print(chartConfig.bottomValue, 2);
        Serial.print("-");
        Serial.print(chartConfig.topValue, 2);
        Serial.print(" | Y-Axis: ");
        Serial.print(chartConfig.topValue, 2);
        Serial.print(",");
        Serial.print(chartConfig.upperMidValue1, 2);
        Serial.print(",");
        Serial.print(chartConfig.upperMidValue2, 2);
        Serial.print(",");
        Serial.print(chartConfig.lowerMidValue1, 2);
        Serial.print(",");
        Serial.print(chartConfig.lowerMidValue2, 2);
        Serial.print(",");
        Serial.print(chartConfig.bottomValue, 2);
        Serial.println();
      }
      break;
    }
  }
}

int scalePriceToChart(float currentPrice) {
  // Map price to 0-240 range (not inverted)
  float scaled = mapFloat(currentPrice,
                        chartConfig.bottomValue,
                        chartConfig.topValue,
                        0.0f,
                        (float)CHART_HEIGHT);
  
  return constrain((int)scaled, 0, CHART_HEIGHT);
}

void updateYAxisLabels() {
  // Update all 7 Y-axis displays (6 divisions + center)
  sendToNextion("n3.val=" + String((int)chartConfig.topValue));       // Top (current + 120)
  sendToNextion("n4.val=" + String((int)chartConfig.upperMidValue1)); // Upper mid 1 (current + 80)
  sendToNextion("n5.val=" + String((int)chartConfig.upperMidValue2)); // Upper mid 2 (current + 40)
  sendToNextion("n7.val=" + String((int)chartConfig.upperMidValue2-40));               // Center (current price)
  sendToNextion("n8.val=" + String((int)chartConfig.lowerMidValue1)); // Lower mid 1 (current - 40)
  sendToNextion("n9.val=" + String((int)chartConfig.lowerMidValue2)); // Lower mid 2 (current - 80)
  sendToNextion("n10.val=" + String((int)chartConfig.bottomValue));   // Bottom (current - 120)
  

}
void printCoinData(uint8_t index) {
  const CoinData &data = coinData[index];
  Serial.println("\nSymbol: " + data.symbol);
  Serial.println("Current Price: " + String(data.currentPrice, 6));
  Serial.println("Price Change (%): " + String(data.priceChangePercent, 2));
  Serial.println("Period High: " + String(data.periodHigh, 6));
  Serial.println("Period Low: " + String(data.periodLow, 6));
  Serial.println("Previous Price: " + String(data.prevPrice, 6));
  Serial.println("-----------------------------");
}

float mapFloat(float x, float in_min, float in_max, float out_min, float out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}